import torch
import typing as th
import functools
import pytorch_lightning as pl
from mdade.made import MADE
from mdade.utils import get_value, process_function_description
from .criterion import MADETrainingCriterion
from .attack import PGDAttacker


class MADETrainer(pl.LightningModule):
    """
    Generic Lightning Module for training MADE models.

    Attributes:
        model: the MADE model (import path)
        criterion: the criterion to use for training
        attack: the attack to use for training
        inputs_transform: the transform to apply to the inputs before forward pass
    """

    def __init__(
        self,
        model_cls: th.Optional[str] = "mdade.MADE",
        model_args: th.Optional[dict] = None,
        criterion_args: th.Optional[dict] = None,
        attack_args: th.Optional[dict] = None,
        inputs_transform: str = None,
        # optimization configs
        optimizer: str = "torch.optim.Adam",
        optimizer_args: th.Optional[dict] = None,
        lr: float = 1e-4,
        scheduler: th.Optional[str] = None,
        scheduler_args: th.Optional[dict] = None,
        scheduler_interval: str = "epoch",
        scheduler_frequency: int = 1,
        scheduler_monitor: th.Optional[str] = None,
    ) -> None:
        """Initialize the trainer.

        Args:
            model_cls: the class of the model to use (import path)
            model_args: the arguments to pass to the model constructor
            criterion_args: the arguments to pass to the criterion constructor (MADETrainingCriterion)
            attack_args: the arguments to pass to the attacker constructor (PGDAttacker)
            inputs_transform:
                the transform function to apply to the inputs before forward pass, can be used for
                applying dequantizations.

        Returns:
            None
        """
        super().__init__()
        self.save_hyperparameters()
        self.model = get_value(self.hparams.model_cls)(**(self.hparams.model_args or dict()))

        # criterion and attacks can be different from the checkpointed model
        self.criterion = MADETrainingCriterion(
            {**(self.hparams.criterion_args or dict()), **(criterion_args or dict())}
        )
        self.attacker = (
            PGDAttacker(
                criterion=self.criterion, **{**(self.hparams.attack_args or dict()), **(attack_args or dict())}
            )
            if (self.hparams.attack_args or attack_args)
            else None
        )
        self.inputs_transform = inputs_transform
        self.lr = self.hparams.lr

    @functools.cached_property
    def inputs_transform_fucntion(self):
        """The transform function (callable) to apply to the inputs before forward pass.

        Returns:
            The compiled callable transform function if `self.inputs_transform` is provided else None
        """
        return (
            process_function_description(self.inputs_transform, entry_function="transform")
            if self.inputs_transform
            else None
        )

    def forward(self, inputs):
        "Placeholder forward pass for the model"
        self.model(inputs)

    def process_inputs(self, batch):
        "Process the inputs before forward pass"
        inputs = batch[0] if isinstance(batch, (tuple, list)) else batch
        inputs = self.inputs_transform_fucntion(inputs) if self.inputs_transform_fucntion else inputs
        if isinstance(self.model, MADE):
            # for mlp models
            inputs = inputs.reshape(inputs.shape[0], -1)

        return inputs

    def step(
        self,
        batch,
        batch_idx: th.Optional[int] = None,
        optimizer_idx: th.Optional[int] = None,
        name: str = "train",
    ):
        """Train or evaluate the model with the given batch.

        Args:
            batch: batch of data to train or evaluate with
            batch_idx: index of the batch
            optimizer_idx: index of the optimizer
            name: name of the step ("train" or "val")

        Returns:
            None if the model is in evaluation mode, else a tensor with the training objective
        """
        is_val = name == "val"
        inputs = self.process_inputs(batch)

        torch.set_grad_enabled(not is_val)
        if self.attacker and not is_val:
            adv_inputs, init_loss, final_loss = self.attacker(model=self.model, inputs=inputs, return_loss=True)
            results = self.criterion(model=self.model, inputs=adv_inputs)
            results["adv/init_loss"] = init_loss
            results["adv/final_loss"] = final_loss
            results["adv/loss_diff"] = final_loss - init_loss
        else:
            results = self.criterion(model=self.model, inputs=inputs)
        for item, value in results.items():
            self.log(
                f"{item}/{name}",
                value.mean(),
                on_step=not is_val,
                on_epoch=is_val,
                logger=True,
                sync_dist=True,
            )
        return results["loss"] if not is_val else None

    def configure_optimizers(self):
        opt_class, opt_dict = torch.optim.Adam, {"lr": self.lr}
        if self.hparams.optimizer:
            opt_class = get_value(self.hparams.optimizer)
            opt_dict = self.hparams.optimizer_args or dict()
            opt_dict["lr"] = self.lr
        opt = opt_class(self.parameters(), **opt_dict)
        if self.hparams.scheduler is None:
            return opt
        sched_class = get_value(self.hparams.scheduler)
        sched_dict = self.hparams.scheduler_args or dict()
        sched = sched_class(opt, **sched_dict)
        sched_instance_dict = dict(
            scheduler=sched,
            interval=self.hparams.scheduler_interval,
            frequency=self.hparams.scheduler_frequency,
            reduce_on_plateau=isinstance(sched, torch.optim.lr_scheduler.ReduceLROnPlateau),
        )
        if self.hparams.scheduler_monitor:
            sched_instance_dict["monitor"] = self.hparams.scheduler_monitor
        return [opt], [sched_instance_dict]

    def training_step(self, batch, batch_idx, optimizer_idx=None):
        "Pytorch Lightning's training_step function"
        return self.step(batch, batch_idx, optimizer_idx, name="train")

    def validation_step(self, batch, batch_idx, optimizer_idx=None):
        "Pytorch Lightning's validation_step function"
        return self.step(batch, batch_idx, optimizer_idx, name="val")



import typing as th
import torch
from .ordered_linear import OrderedLinear
from .ordered_residual import OrderedResidual1D
from ..utils import get_value


class OrderedBlock(torch.nn.Module):
    def __init__(
        self,
        # linear args
        in_features: int,
        out_features: int,
        bias: bool = True,
        # activation
        activation: th.Optional[str] = "torch.nn.GELU",
        activation_args: th.Optional[dict] = None,
        # batch norm
        batch_norm: bool = True,
        batch_norm_args: th.Optional[dict] = None,
        # residual
        residual: bool = True,
        residual_scale: bool = True,
        # ordering args
        auto_connection: bool = True,
        # general parameters
        device: th.Optional[torch.device] = None,
        dtype: th.Optional[torch.dtype] = None,
    ):
        super().__init__()
        self.linear = OrderedLinear(
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            auto_connection=auto_connection,
            device=device,
            dtype=dtype,
        )
        self.activation = get_value(activation)(**(activation_args or dict())) if activation else None
        self.batch_norm = (
            torch.nn.BatchNorm1d(num_features=out_features, dtype=dtype, device=device, **(batch_norm_args or dict()))
            if batch_norm
            else None
        )
        self.residual = (
            OrderedResidual1D(
                in_features=in_features,
                out_features=out_features,
                auto_connection=auto_connection,
                device=device,
                scale=residual_scale,
            )
            if residual
            else None
        )

    @property
    def ordering(self):
        return self.linear.ordering

    def reorder(
        self,
        inputs_ordering: torch.IntTensor,
        ordering: th.Optional[torch.IntTensor] = None,
        allow_detached_neurons: bool = True,
        highest_ordering_label: th.Optional[int] = None,
        generator: th.Optional[torch.Generator] = None,
    ):
        self.linear.reorder(
            inputs_ordering=inputs_ordering,
            ordering=ordering,
            allow_detached_neurons=allow_detached_neurons,
            highest_ordering_label=highest_ordering_label,
            generator=generator,
        )

        if self.residual:
            self.residual.reorder(
                ordering=self.linear.ordering,
                inputs_ordering=inputs_ordering,
            )

    def forward(self, inputs):
        outputs = self.linear(inputs)
        outputs = self.activation(outputs) if self.activation else outputs
        outputs = self.batch_norm(outputs) if self.batch_norm else outputs
        outputs = self.residual(inputs, outputs) if self.residual else outputs
        return outputs



import torch
import typing as th


def freeze_params(model: torch.nn.Module):
    old_states = []
    for param in model.parameters():
        old_states.append(param.requires_grad)
        param.requires_grad = False
    return old_states


def unfreeze_params(
    model: torch.nn.Module, old_states: th.Optional[th.List[bool]] = None
):
    for idx, param in enumerate(model.parameters()):
        param.requires_grad = True if old_states is None else old_states[idx]




import torch
import functools
import typing as th
from mdade.utils import process_function_description
from .utils import freeze_params, unfreeze_params


class PGDAttacker:
    def __init__(
        self,
        criterion,
        criterion_roi: th.Union[str, th.List[str]] = "loss",
        num_iters: int = 1,
        alpha: float = 0.01,
        epsilon: th.Optional[float] = None,
        p_norm: str = "inf",
        inputs_clamp="lambda x: torch.clamp(x, 0., 1.)",
        random_start: bool = True,
        random_start_generation: str = "lambda x: (torch.rand_like(x) - 0.5) * 2",
    ):
        self.criterion = criterion
        self.criterion_roi = criterion_roi
        self.num_iters = num_iters
        self.epsilon = epsilon
        self.alpha = alpha
        self.random_start = random_start
        self.p_norm = p_norm
        self.random_start_generation_function = random_start_generation
        self.inputs_clamp_function = inputs_clamp

    @functools.cached_property
    def generate_random_start(self):
        return process_function_description(self.random_start_generation_function, entry_function="process")

    @functools.cached_property
    def clamp_inputs(self):
        return process_function_description(self.inputs_clamp_function, entry_function="process")

    def objective(self, model, inputs):
        criterion_results = self.criterion(model=model, inputs=inputs)
        if isinstance(self.criterion_roi, str):
            return criterion_results[self.criterion_roi].mean()
        return sum(criterion_results[name] for name in self.criterion_roi).mean()

    @classmethod
    def renorm_adversary(
        cls,
        adv_inputs,
        epsilon: th.Optional[float] = None,
        p_norm: th.Optional[th.Union[str, int]] = "inf",
    ):
        if epsilon is None:
            return adv_inputs
        if p_norm == "inf":
            return adv_inputs.clamp(-epsilon, epsilon)
        norms = torch.norm(
            adv_inputs,
            dim=torch.arange(len(adv_inputs.shape))[1:].tolist(),
            p=p_norm,
            keepdim=True,
        )
        return torch.where(norms > epsilon, adv_inputs * epsilon / norms, adv_inputs)

    def __call__(self, model, inputs, return_loss=True, force_eval: bool = True):
        torch.set_grad_enabled(True)
        if self.random_start:
            delta = self.generate_random_start(inputs)
            delta.requires_grad = True
        else:
            delta = torch.zeros_like(inputs, requires_grad=True)

        # freezing model params
        params_state = freeze_params(model)
        # force eval
        model_training = model.training
        if force_eval:
            model.eval()

        for i in range(self.num_iters):
            loss = self.objective(model=model, inputs=inputs.detach() + delta)
            if i == 0 and return_loss:
                init_loss = loss.detach()
            loss.backward()
            delta.data.copy_(delta.data + self.alpha * delta.grad.detach().sign())
            delta.data.copy_(self.renorm_adversary(adv_inputs=delta, epsilon=self.epsilon, p_norm=self.p_norm))
            delta.grad.zero_()
        if return_loss:
            final_loss = self.objective(model=model, inputs=inputs.detach() + delta).detach()

        # unfreezing model
        unfreeze_params(model, params_state)
        # unforce eval
        if model_training:
            model.train()
        if return_loss:
            return (
                self.clamp_inputs(inputs.detach() + delta.detach()),
                init_loss,
                final_loss,
            )
        return self.clamp_inputs(inputs.detach() + delta.detach())



import typing as th
from numpy import isin
from torchvision import transforms as tv_transforms
from mdade.utils import get_value, process_function_description
import inspect


def initialize_transforms(transforms: th.Optional[th.Union[list, dict, th.Any]]):
    if transforms is None or (not inspect.isclass(transforms) and callable(transforms)):
        return transforms

    # list of other transforms
    if isinstance(transforms, list):
        return tv_transforms.Compose([initialize_transforms(i) for i in transforms])

    # either a class and args, or a code block and entry function
    if isinstance(transforms, dict):
        if "cls" in transforms:
            return get_value(transforms["cls"])(transforms.get("args", dict()))
        value = process_function_description(transforms, "transform")
        return value() if inspect.isclass(value) else value
    if isinstance(transforms, str):
        try:
            return get_value(transforms)()
        except:
            return process_function_description(transforms, "transform")



import torch
import typing as th
import functools
import inspect
from .ordered_linear import OrderedLinear
from ..utils import process_function_description, get_value


class AutoRegressiveDensityEstimator1D(OrderedLinear):
    def __init__(
        self,
        dims_count: int,
        in_features: int,
        bias: bool = True,
        num_mixtures: int = 1,
        share_params_features: th.Optional[int] = None,
        distribution: str = "torch.distributions.Normal",
        distribution_args: th.Optional[dict] = None,
        distribution_params_transforms: th.Optional[dict] = None,
        device: th.Optional[torch.device] = None,
        dtype: th.Optional[torch.dtype] = None,
    ):
        self.num_mixtures = num_mixtures
        self.__distribution_name = distribution
        self.distribution = get_value(distribution) if isinstance(distribution, str) else distribution
        assert issubclass(
            self.distribution, torch.distributions.Distribution
        ), "expected a subclass of `torch.distributions.Distribution` as the estimator's distribution"
        self.distribution = functools.partial(self.distribution, **(distribution_args or dict()))
        self.share_params_features = share_params_features
        self.dims_count = dims_count
        super().__init__(
            in_features=in_features,
            out_features=(
                dims_count
                * (len(self.distribution_params_names) + (1 if self.num_mixtures > 1 else 0))
                * self.num_mixtures
            )
            if share_params_features is None
            else (share_params_features * dims_count),
            bias=bias,
            device=device,
            dtype=dtype,
            auto_connection=False,
        )

        # shared linear layer to compute dims' distribution parameters from features
        self.params_compute = (
            torch.nn.Linear(
                in_features=share_params_features,
                out_features=(len(self.distribution_params_names) + (1 if self.num_mixtures > 1 else 0))
                * self.num_mixtures,
                bias=bias,
                device=device,
                dtype=dtype,
            )
            if share_params_features is not None
            else None
        )
        # outsourced functions to transform distribution parameters
        self.distribution_params_transforms = distribution_params_transforms

    @functools.cached_property
    def distribution_params_transforms_functions(self):
        return {
            name: process_function_description(value, "transform")
            for name, value in (self.distribution_params_transforms or dict()).items()
        }

    @functools.cached_property
    def distribution_params_names(self):
        params = inspect.signature(self.distribution).parameters
        return [
            name
            for name in params
            if name not in ["probs", "validate_args"]
            and (params[name].default is None or params[name].default is inspect._empty)
        ]

    @functools.cached_property
    def parameter_indeces(self):
        result = torch.concat(
            [
                (
                    (
                        torch.arange(self.dims_count, dtype=torch.long)
                        * (len(self.distribution_params_names) + (1 if self.num_mixtures > 1 else 0))
                        * self.num_mixtures
                    )[:, None]
                    + torch.arange(self.num_mixtures, dtype=torch.long)
                    + (index * self.num_mixtures)
                ).unsqueeze(0)
                for index in range(len(self.distribution_params_names) + (1 if self.num_mixtures > 1 else 0))
            ],
            dim=0,
        )
        return result if self.num_mixtures > 1 else result.squeeze(-1)

    def transform_distribution_parameters(self, params_logits) -> th.Dict[str, torch.Tensor]:
        "transforms distribution parameters given their raw logits (output of model)"
        results = {
            name: (
                params_logits[:, i]
                if name not in self.distribution_params_transforms_functions
                else self.distribution_params_transforms_functions[name](params_logits[:, i])
            )
            for i, name in enumerate(self.distribution_params_names)
        }
        if self.num_mixtures > 1:
            results["mixture_logits"] = (
                self.distribution_params_transforms_functions["mixture_logits"](params_logits[:, -1])
                if "mixture_logits" in self.distribution_params_transforms_functions
                else params_logits[:, -1]
            )
        return results

    def distributions(self, params_logits, params=None) -> torch.distributions.Distribution:
        params = params or self.transform_distribution_parameters(params_logits)
        component_params = {key: value for key, value in params.items() if key != "mixture_logits"}
        component_distributions = self.distribution(**(component_params))
        if self.num_mixtures == 1:
            return component_distributions
        mixture_distributions = torch.distributions.Categorical(logits=(params["mixture_logits"]))
        return torch.distributions.MixtureSameFamily(mixture_distributions, component_distributions)

    def log_prob(self, inputs, params_logits, params=None, reduce=False) -> torch.Tensor:
        log_probs = self.distributions(params_logits=params_logits, params=params).log_prob(inputs)
        return log_probs.sum(-1) if reduce else log_probs

    def forward(self, input_features) -> torch.Tensor:
        "Computes the untransformed parameter logits, given ordered input_features"
        if self.params_compute is None:
            return super().forward(input_features)[:, self.parameter_indeces]

        features = super().forward(input_features).reshape(input_features.shape[0], self.dims_count, -1)
        return self.params_compute(features).reshape(input_features.shape[0], -1)[:, self.parameter_indeces]

    def extra_repr(self):
        return "distribution={}, num_mixtures={}{}".format(
            self.__distribution_name,
            self.num_mixtures,
            f", params_features={self.share_params_features}" if self.share_params_features else "",
        )



import torch
import typing as th
import functools


class OrderedLayerMixin1D:
    """
    Mixin to support ordered 1D outputs that maintain an autoregressive data flow.

    This class should be initialized after `torch.nn.Module` is inherited and
    initialized in the desired module.

    Attributes:
        ordering: Current ordering of the output neurons.
        auto_connection: Whether to allow equal label connections.
        mask:
            A zero-one 2D mask [OUTPUT x INPUT] matrix defining the connectivity
            of output and input neurons.

    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        auto_connection: bool = True,
        device: th.Optional[torch.device] = None,
        mask_dtype: torch.dtype = torch.uint8,
    ) -> None:
        """
        Initiates ordering related buffers.

        Args:
            in_featuers: Number of input dimensions with ordering.
            out_features: Number of output dimensions with ordering.
            auto_connection:
                A boolean value specifying whether output neurons with the
                same labels as the input neurons can be connected.
            device: The device to instanciate the buffers on.
            masked_dtype: Data type for mask matrix.
        Retures:
            None
        """
        self.auto_connection = auto_connection
        self.register_buffer("ordering", torch.empty(out_features, device=device, dtype=torch.int))
        self.register_buffer(
            "mask",
            torch.ones(out_features, in_features, device=device, dtype=mask_dtype),
        )

    @functools.cached_property
    def connection_operator(self):
        return torch.less_equal if self.auto_connection else torch.less

    def reorder(
        self,
        inputs_ordering: torch.IntTensor,
        ordering: th.Optional[torch.IntTensor] = None,
        allow_detached_neurons: bool = True,
        highest_ordering_label: th.Optional[int] = None,
        generator: th.Optional[torch.Generator] = None,
    ) -> None:
        """
        (Re)computes the output ordering and the mask matrix.

        This function computes the layer's ordering based on the given
        ordering (to be enforced) or the ordering of layer's inputs and a
        random number generator. Optionally you can choose to disallow
        detached neurons, so that the layer ordering labels are chosen
        from higher values than the last layer.

        Args:
            inputs_ordering:
                Ordering of inputs to the layer used for computing new
                layer inputs (randomly out with the generator), and the
                connectivity mask.
            ordering: An optional ordering to be enforced
            allow_detached_neurons:
                If true, the minimum label for this layer's outputs would
                start from zero regardless of whether the previous layer's
                ordering. Else, layer's labels will start from the minimum
                label of inputs.
            highest_ordering_label: Maximum label to use.
            generator: Random number generator.

        Returns:
            None
        """
        if ordering is not None:
            # enforcing ordering and if needed repeating the providded ordering
            # across layer's ordering. (especially used for predicting autoregressive
            # mixture of parameters which requires an output with a size with a multiple
            # of the number of input dimensions.)
            self.ordering.data.copy_(torch.repeat_interleave(ordering, self.ordering.shape[0] // ordering.shape[0]))
        else:
            self.ordering.data.copy_(
                torch.randint(
                    low=0 if allow_detached_neurons else inputs_ordering.min(),
                    high=highest_ordering_label or inputs_ordering.max(),
                    size=self.ordering.shape,
                    generator=generator,
                )
            )
        self.mask.data.copy_(self.connection_operator(inputs_ordering[:, None], self.ordering[None, :]).T)



import typing as th
import torch
import functools
from mdade.utils import process_function_description, safe_function_call_wrapper
from mdade.made import MADE

# types
FactorsDict = th.Dict[str, torch.Tensor]
ResultsDict = th.Dict[str, torch.Tensor]


class MADETrainingCriterion:
    """Training objective for Masked Autoregressive Models.

    Attributes:
        params_regularizations (dict):
            A dictionary of regularization functions for the parameters of the model.
        params_regularizations_factors (dict):
            A dictionary of factors for the regularization functions for the parameters of the model.
        scale_regularizations (bool):
            Whether to scale the regularization losses to be in the same range as the negative log-likelihood.
    """

    def __init__(
        self,
        params_regularizations: th.Optional[th.Dict[str, str]] = None,
        params_regularizations_factors: th.Optional[th.Dict[str, str]] = None,
        scale_regularizations: bool = False,
    ):
        self.params_regularizations = params_regularizations or dict()
        self.params_regularizations_factors = params_regularizations_factors or dict()
        self.scale_regularizations = scale_regularizations

    @functools.cached_property
    def params_regularization_factor_functions(self):
        return {
            name: safe_function_call_wrapper(process_function_description(value, entry_function="factor"))
            for name, value in self.params_regularizations_factors.items()
        }

    @functools.cached_property
    def params_regularization_functions(self):
        return {
            name: safe_function_call_wrapper(process_function_description(value, entry_function="regularize"))
            for name, value in self.params_regularizations.items()
        }

    def __call__(
        self, model: MADE, inputs, trainer=None, return_factors=False
    ) -> th.Union[ResultsDict, th.Tuple[ResultsDict, FactorsDict]]:
        if not self.params_regularization_functions:
            # shortcut for trainings with no parameter regularizations
            nll = -model.log_prob(inputs, reduce=True).mean()
            return dict(loss=nll, nll=nll) if not return_factors else (dict(loss=nll, nll=nll), dict())

        results = dict(nll=0.0, loss=0.0)
        regularizations = {name: 0.0 for name in self.params_regularization}
        for i in range(model.num_masks or 1):
            model.reorder()
            params_logits = model(inputs)
            params = model.density_estimator.transform_distribution_parameters(params_logits)
            for name, func in self.params_regularization_functions.items():
                regularizations[name] += func(
                    params[name], trainer=trainer, model=model
                ).mean()  # mean will ensure that regularization results are scalars
            results["nll"] -= model.density_estimator.log_prob(
                inputs,
                params_logits=params_logits,
                params=params,
                reduce=True,
            ).mean()

        # normalizing by the number of masks
        factors = dict()
        results["nll"] /= model.num_masks or 1
        results["loss"] += results["nll"]
        for name in regularizations:
            regularizations[name] /= model.num_masks or 1
            results[f"params_regularization/{name}"] = regularizations[name]

            factor = self.params_regularization_factor.get(name, 1.0)
            factors[name] = factor(regularizations[name], trainer=trainer, model=model) if callable(factor) else factor

            results["loss"] += (
                (regularizations[name] / regularizations[name].data.abs() * results["nll"].data)
                if self.scale_regularizations
                else regularizations[name]
            ) * factors[name]
        return results if not return_factors else results, factors

    def __repr__(self):
        regularizations = f"params_regularization={self.params_regularizations}" if self.params_regularizations else ""
        regularization_factors = (
            f", params_regularization_factor={self.params_regularizations_factors}"
            if self.params_regularizations_factors
            else ""
        )
        return f"{self.__class__.__name__}({regularizations}{regularization_factors})"



import importlib
import typing as th
import types
import inspect

# for eval context
import torch



def safe_function_call_wrapper(function: th.Callable):
    signature = inspect.signature(function)
    params = signature.parameters

    def wrapper(*args, **kwargs):
        call_kwargs = {name: kwargs[name] for name in params if name in kwargs}
        return function(*args, **call_kwargs)

    return wrapper


def generate_function(code_block, function: str) -> th.Callable[[th.Any], th.Any]:
    context = dict()
    exec(code_block, dict(), context)
    return types.FunctionType(
        code=context[function].__code__,
        globals=context,
        name=function,
        argdefs=context[function].__defaults__,
    )


def process_function_description(
    function: th.Union[th.Callable, str, th.Dict[str, str]], entry_function
) -> th.Callable:
    if callable(function) or not isinstance(function, (str, dict)):
        return function
    try:
        return eval(function if isinstance(function, str) else function["code"])
    except SyntaxError:
        return generate_function(
            code_block=function if isinstance(function, str) else function["code"],
            function=function.get("entry", entry_function) if isinstance(function, dict) else entry_function,
        )


def import_context(name: str):
    return importlib.import_module(name)


def get_value(name: str, context: th.Optional[th.Any] = None, strict: bool = True):
    var = context if context is not None else import_context(name.split(".")[0])
    for split in name.split(".")[(0 if context is not None else 1) :]:
        if isinstance(var, dict):
            if split not in var:
                if strict:
                    raise KeyError('Invalid key "%s"' % name)
                else:
                    return None
            var = var[split]
        else:
            if not hasattr(var, split):
                if strict:
                    raise AttributeError("Invalid attribute %s" % name)
                else:
                    return None
            var = getattr(var, split)
    return var



import torch
import typing as th
from .ordering_mixin import OrderedLayerMixin1D


class OrderedLinear(
    torch.nn.Linear,
    OrderedLayerMixin1D,
):
    """
    Linear layer with ordered outputs to maintain an autoregressive data flow.

    Attributes:
        ordering:
            Current ordering of the output neurons.
        mask:
            A zero-one 2D mask matrix defining the connectivity of output and
            input neurons.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        device: th.Optional[torch.device] = None,
        dtype: th.Optional[torch.dtype] = None,
        auto_connection: bool = True,
    ) -> None:
        """
        Initializes module by initializing an ordering mixin and a linear layer.

        Args:
            in_features: Number of input feature dimensions in 1D.
            out_features: Number of output feature dimensions in 1D.
            bias: Whether to use a bias vector for the linear operation.
            device: The destination device for the buffers and parameters.
            dtype: Data type for linear layer parameters.
            masked_dtype: Data type for mask matrix.
            auto_connection: Whether to allow equal label connections.

        Returns:
            None
        """
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            device=device,
            dtype=dtype,
        )
        OrderedLayerMixin1D.__init__(
            self,
            in_features=in_features,
            out_features=out_features,
            device=device,
            auto_connection=auto_connection,
        )

    def forward(self, inputs: torch.Tensor):
        """
        Computes masked linear operation.

        Args:
            inputs: Input tensor (batched in the first dimensions.)

        Returns:
            A `torch.Tensor` which equals to masked linear operation on inputs, or:
                `inputs @ (self.mask * self.weights).T + self.bias`
        """
        return torch.nn.functional.linear(inputs, self.mask * self.weight, self.bias)



import typing as th
import torch
import torchvision
import pytorch_lightning as pl


class CheckOutlierCallback(pl.Callback):
    """Evaluate the model by recording the nll assigned to random noise.

    If the training loss stays constant across iterations, this score is likely to show
     the progress of the model in detecting "outliers".

    Attributes:
        batch_size (int): Number of random samples to check
        every_n_epochs (int): Check every n epochs
    """

    def __init__(self, batch_size: int = 32, every_n_epochs: int = 1):
        super().__init__()
        self.batch_size = batch_size
        self.every_n_epochs = every_n_epochs

    def compute(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        with torch.no_grad():
            pl_module.eval()
            # todo: support non-mlp based models
            rand_inputs = torch.rand(self.batch_size, pl_module.model.in_features, device=pl_module.device)
            rand_out = pl_module.criterion(model=pl_module.model, inputs=rand_inputs)
            pl_module.train()

        trainer.logger.experiment.add_scalar("rand_nll", rand_out, global_step=trainer.global_step)

    def on_validation_epoch_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if trainer.current_epoch % self.every_n_epochs:
            return
        return self.compute(trainer=trainer, pl_module=pl_module)


class SampleAdversariesCallback(pl.Callback):
    """Evaluate the model by sampling adversarial examples.

    If the training loss stays constant across iterations, this score is likely to show
     the progress of the model in detecting "outliers".

    Attributes:
        dataloader (str): dataloader (val/train/test) to use for initial inputs
        dataset_args (dict): arguments to pass to the dataloader (if to be different than the original)
        reshape (tuple): reshape the inputs to this shape [H, W]
        difference_map_overlay_alpha (float):
            alpha value for the difference map overlay (0 not to log difference maps)
        difference_map_normalize (bool): normalize the difference map values to [-1, 1]
        adversaries_grid_args (dict): arguments to pass to the grid visualization of the adversaries
        difference_map_grid_args (dict): arguments to pass to the grid visualization of the difference maps
    """

    def __init__(
        self,
        every_n_epochs: int = 1,
        dataloader="train",
        dataloader_args: th.Optional[dict] = None,
        reshape: th.Optional[th.Union[list, th.Tuple[int, int]]] = None,
        difference_map_overlay_alpha: float = 0.6,
        difference_map_normalize: bool = False,
        grid_args: th.Optional[dict] = None,
        adversaries_grid_args: th.Optional[dict] = None,
        difference_map_grid_args: th.Optional[dict] = None,
    ):
        super().__init__()

        # interval settings
        self.every_n_epochs = every_n_epochs

        # dataloader attributes
        self.dataloader_name = dataloader
        self.dataloder = None  # will be added in setup stage
        self.dataloader_args = dataloader_args or dict()

        # visualization configs
        self.reshape = tuple(reshape) if reshape else None
        self.difference_map_overlay_alpha = difference_map_overlay_alpha
        self.adversaries_grid_args = {**(grid_args or dict()), **(adversaries_grid_args or dict())}
        self.difference_map_grid_args = {**(grid_args or dict()), **(difference_map_grid_args or dict())}
        self.difference_map_normalize = difference_map_normalize

    def setup(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", stage: th.Optional[str] = None) -> None:
        if stage == "fit":
            self.dataloader = getattr(trainer.datamodule, f"{self.dataloader_name}_dataloader")(**self.dataloader_args)
            self.inputs_iterator = iter(self.get_inputs_generator(trainer=trainer, pl_module=pl_module))

    def process_shape(self, inputs, force_grayscale=True, three_channel_grayscales=True):
        """Process the shape of the inputs to be used in the visualization.

        If the input tensors are reshaped image tensors, they shall be reshaped to the provided width and height.
        Else, if the input tensors represent non-single channel images, they shall be converted to grayscale.
        """
        results = inputs.reshape(inputs.shape[0], -1, *self.reshape) if self.reshape else inputs
        if results.shape[1] == 1:
            results = results.repeat(1, 3 if three_channel_grayscales else 1, 1, 1)
        elif force_grayscale or (results.shape[1] > 3 or results.shape[1] == 2):
            results = results.mean(1, keepdim=True)  # convert to grayscale
            results = results.repeat(1, 3 if three_channel_grayscales else 1, 1, 1)
        return results

    def process_difference_map(
        self, adversarial_inputs, inputs, overlay_alpha: th.Optional[float] = None, normalize: th.Optional[bool] = None
    ):
        # config
        overlay_alpha = overlay_alpha if overlay_alpha is not None else self.difference_map_overlay_alpha
        normalize = normalize if normalize is not None else self.difference_map_normalize

        # process shapes
        difference_map = self.process_shape(adversarial_inputs - inputs, three_channel_grayscales=False)
        inputs = self.process_shape(inputs)
        # compute difference maps (positive changes are red, negative changes are blue)
        changes = torch.zeros_like(inputs)
        changes[:, 0] = torch.where(difference_map > 0, difference_map, 0)[:, 0]  # red
        changes[:, 2] = torch.where(difference_map < 0, -difference_map, 0)[:, 0]  # blue
        if normalize:
            changes /= changes.abs().max()
        return inputs * (1 - overlay_alpha) + overlay_alpha * changes

    def get_inputs_generator(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> torch.Tensor:
        """Retrieve and process a batch of inputs from the dataloader.

        Returns:
            Tensor of inputs
        """
        while True:
            for batch in self.dataloader:
                results = pl_module.process_inputs(batch).to(pl_module.device)
                yield results

    def get_adversaries(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule, inputs: torch.Tensor
    ) -> torch.Tensor:
        """Get adversaries to the model

        Arguments:
            trainer: trainer
            pl_module: pl_module
            inputs: inputs to the model

        Returns:
            Tensor of adversarial inputs
        """
        return pl_module.attacker(inputs=inputs, model=pl_module.model, return_loss=False)

    def on_train_epoch_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if (
            trainer.current_epoch % self.every_n_epochs  # interval
            or not hasattr(pl_module, "attacker")  # no attacker for this model
            or pl_module.attacker is None  # attacker is not set
        ):
            return

        inputs = next(self.inputs_iterator)
        adversaries = self.get_adversaries(trainer=trainer, pl_module=pl_module, inputs=inputs)

        # visualize adversaries
        trainer.logger.experiment.add_image(
            f"attack/adversaries",
            torchvision.utils.make_grid(self.process_shape(adversaries), **self.adversaries_grid_args),
            global_step=trainer.global_step,
        )

        # visualize difference maps
        if self.difference_map_overlay_alpha:
            pl_module.logger.experiment.add_image(
                "attack/difference_map",
                torchvision.utils.make_grid(
                    self.process_difference_map(adversarial_inputs=adversaries, inputs=inputs),
                    **self.difference_map_grid_args,
                ),
                global_step=trainer.global_step,
            )


class SampleCallback(pl.Callback):
    """
    Samples generations from the model and save them to the experiment log.

    Attributes:
        num_samples (int): Number of images to generate
        every_n_epochs (int): Only save those images every N epochs (otherwise logs gets quite large)
        reshape: optionally reshape model samples (for mlp-based models)
        grid_args: arguments to pass to torchvision.utils.make_grid
        mask_index: index of the mask to use for sampling from model (-1 for all masks)
    """

    def __init__(
        self,
        num_samples: int = 8,
        every_n_epochs: int = 1,
        reshape: th.Optional[th.Union[th.Union[int, int], list]] = None,
        grid_args: th.Optional[dict] = None,
        mask_index: th.Optional[int] = 0,
    ):
        super().__init__()
        self.num_samples = num_samples
        self.every_n_epochs = every_n_epochs
        self.reshape = reshape
        self.mask_index = mask_index
        self.grid_args = grid_args or dict()

    def log_samples(self, trainer: pl.Trainer, name: str, samples):
        trainer.logger.experiment.add_image(
            name,
            torchvision.utils.make_grid(samples, **self.grid_args),
            global_step=trainer.global_step,
        )

    def get_samples(self, pl_module: pl.LightningModule, mask_index: th.Optional[int] = None):
        samples = pl_module.model.sample(num_samples=self.num_samples, mask_index=mask_index)
        if self.reshape is not None:
            samples = samples.reshape(samples.shape[0], *self.reshape)
        return samples

    def on_train_epoch_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule):
        """Samples generations from the model and save them to the experiment log."""
        if trainer.current_epoch % self.every_n_epochs:  # interval check
            return

        if self.mask_index == -1 and pl_module.model.num_masks > 1:
            # sample all masks
            for i in range(pl_module.model.num_masks):
                self.log_samples(
                    trainer=trainer,
                    name=f"samples/mask/{i}",
                    samples=self.get_samples(pl_module=pl_module, mask_index=i),
                )
            return
        self.log_samples(
            trainer=trainer,
            name=f"samples/mask/{self.mask_index}",
            samples=self.get_samples(pl_module=pl_module, mask_index=self.mask_index),
        )



import torch
import typing as th
import functools
from .utils import process_function_description
from .layers import OrderedBlock, AutoRegressiveDensityEstimator1D


class MADE(torch.nn.Module):
    """
    Masked Autoregressive (Auto-encoder) Density Estimator (MADE)

    Attributes:
        in_features: The number of input features.
        num_masks: The number of masks to use. (0 for infinite masks)
        layers: Hidden layers of the MADE. (list of number of features)
        bias: If True, adds a bias term to linear computations.
        residual: If True, adds a residual connections.
        residual_scale: If True, scales the residual connection by the number of connections.
        activation: The activation function to use (None for linear model).
        activation_args: The arguments to pass to the activation function.
        batch_norm: If True, adds a batch normalization layer in blocks.
        batch_norm_args: The arguments to pass to the batch normalization layers.
        num_mixtures: The number of mixtures to use in the density estimator.
        share_params_features: The number of features to share the parameters of the density estimator.
        distribution:
            The distribution to use in the density estimator. (subclass of torch.distributions.Distribution)
        distribution_args: The arguments to pass to the distribution class.
        distribution_params_transforms: The transforms to apply to the parameters of the distribution.
        bias_distribution: If True, adds bias terms to the density estimator.
        seed: The seed to use for the random number generator.
        masks_kind: The kind of masks and orderings to use. ("random" or "repeat") (default: random)
        device: The device to use.
        dtype: The data type to use.
        safe_grad_hook: The function to use to hook the gradients (right after forward).
    """

    def __init__(
        self,
        in_features: int,
        num_masks: int,
        # blocks
        layers: th.List[int],
        bias: bool = True,
        residual: bool = False,
        residual_scale: bool = True,
        activation: th.Optional[str] = "torch.nn.GELU",
        activation_args: th.Optional[dict] = None,
        batch_norm: bool = True,
        batch_norm_args: th.Optional[dict] = None,
        # distribution
        num_mixtures: int = 1,
        share_params_features: th.Optional[int] = None,
        distribution: str = "torch.distributions.Normal",
        distribution_args: th.Optional[dict] = None,
        distribution_params_transforms: th.Optional[dict] = None,
        bias_distribution: bool = True,
        # general
        seed: int = 0,
        masks_kind: str = "random",
        device: th.Optional[torch.device] = None,
        dtype: th.Optional[torch.dtype] = None,
        # grad safety
        safe_grad_hook: str = "lambda grad: torch.where(torch.isnan(grad) + torch.isinf(grad), torch.zeros_like(grad), grad)",
    ):
        super().__init__()
        self.in_features = in_features
        self.num_masks = num_masks
        self.masks_kind = masks_kind
        self.__mask_indicator = 0
        self.seed = seed
        self.current_seed = seed
        self.generator = torch.Generator().manual_seed(seed)
        self.residual = residual
        self.layers = torch.nn.Sequential(
            *[
                OrderedBlock(
                    in_features=in_features if not i else layers[i - 1],
                    out_features=layers[i],
                    bias=bias,
                    residual=residual,
                    residual_scale=residual_scale,
                    activation=activation,
                    activation_args=activation_args,
                    batch_norm=batch_norm,
                    batch_norm_args=batch_norm_args,
                    device=device,
                    dtype=dtype,
                )
                for i in range(len(layers))
            ]
        )
        self.density_estimator = AutoRegressiveDensityEstimator1D(
            dims_count=in_features,
            in_features=layers[-1],
            bias=bias_distribution,
            num_mixtures=num_mixtures,
            share_params_features=share_params_features,
            distribution=distribution,
            distribution_args=distribution_args,
            distribution_params_transforms=distribution_params_transforms,
            device=device,
            dtype=dtype,
        )

        self.register_buffer("ordering", torch.arange(self.in_features, dtype=torch.int, device=device))
        self.safe_grad_hook = safe_grad_hook

        self.reorder(initialization=True)

    @functools.cached_property
    def safe_grad_hook_function(self):
        return process_function_description(self.safe_grad_hook, entry_function="hook")

    def check_autoregressive_property(self, missing_deps: bool = True, extra_deps: bool = True, mask_index: int = 0):
        """
        Checks if the MADE is autoregressive.

        Args:
            missing_deps: If True, checks if the MADE is missing dependencies.
            extra_deps: If True, checks if the MADE has extra dependencies.
            mask_index: The index of the mask to use.

        Returns:
            None
        """
        inputs = torch.ones(1, self.in_features)
        inputs.requires_grad = True

        # force evaluation mode
        model_training = self.training
        self.eval()

        # reorder to specified mask_index
        current_mask = self.__mask_indicator
        self.reorder(mask_index=mask_index)
        outputs = self.log_prob(inputs, mask_index=mask_index)[0]
        outputs = outputs.squeeze(0)
        inputs.reshape(-1)

        for idx in range(outputs.shape[0]):
            # backpropogate to inputs to find data flow dependencies
            label = self.ordering[idx].item()
            outputs[idx].backward(retain_graph=True)  # reuse graph to avoid O(n^2) forward passes
            deps_idxs = torch.where(inputs.grad[0])[0]  # get data flow dependencies (non-zero gradients)
            deps_labels = set(self.ordering[deps_idxs].tolist())  # get labels of dependencies

            if missing_deps:  # check for missing dependencies (i.e. labels smaller or equal to currentlabel)
                missing = [i for i in range(label + 1) if i not in deps_labels]
                if missing:
                    print(f"output({idx})[label={label}] -> not depended on labels: {missing}")
            if extra_deps:  # check for extra dependencies (larger than current label)
                extra = [i for i in range(label + 1, self.in_features) if i in deps_labels]
                if extra:
                    print(f"output({idx})[label={label}] -> extra dependancies on labels: {extra}")
            inputs.grad.zero_()  # restart gradient calculation

        # resotre training mode
        if model_training:
            self.train()
        # restore ordering
        self.reorder(current_mask)

    def reorder(self, mask_index=None, initialization=False):
        """
        Either reorders to the next mask or to the specified mask.

        Args:
            mask_index: The index of the mask to use.
            initialization: If True, the mask is being initialized for the first time.

        Returns:
            None
        """
        # if there is only one mask, do nothing except if initialization is True (to initialize layers' orderings)
        if self.num_masks == 1 and not initialization:
            return

        # if mask_index is None, use next mask else use specified mask (and remember the original ordering)
        if self.__mask_indicator or mask_index is not None:
            # each ordering corresponds to a random number generotor and a random seed
            self.generator = torch.Generator().manual_seed(
                self.seed + (mask_index if mask_index is not None else self.__mask_indicator)
            )
            if mask_index is not None:  # if mask_index is not None, we are initializing the mask[mask_index]
                self.__mask_indicator = mask_index

        # initialize inputs ordering (mask_index=0 is cannonical ordering the rest are random permutations)
        if not self.num_masks or self.__mask_indicator:
            self.ordering.data.copy_(torch.randperm(self.in_features, generator=self.generator))
        elif not self.__mask_indicator:
            self.ordering.data.copy_(torch.arange(self.in_features, dtype=torch.int, device=self.ordering.device))
            self.generator = torch.Generator().manual_seed(self.seed)

        for i, layer in enumerate(self.layers):
            layer.reorder(
                inputs_ordering=self.ordering if not i else self.layers[i - 1].ordering,
                generator=self.generator,
                # force ordering to be the same for all layers if masks_kind is repeat
                ordering=self.ordering if self.masks_kind == "repeat" else None,
                allow_detached_neurons=self.residual,
                highest_ordering_label=self.in_features,
            )
        self.density_estimator.reorder(inputs_ordering=self.layers[-1].ordering, ordering=self.ordering)

        # if mask_index is None, move to next mask for the next model.reorder() call
        if mask_index is None and not initialization:
            self.__mask_indicator = ((self.__mask_indicator + 1) % self.num_masks) if self.num_masks else 0

    def forward(self, inputs, mask_index: th.Optional[int] = None, safe_grad: bool = True):
        if mask_index is None:
            results = self.density_estimator(self.layers(inputs))
            if results.requires_grad and safe_grad:
                results.register_hook(self.safe_grad_hook_function)
            return results

    def distributions(self, inputs):
        return self.density_estimator.distributions(self(inputs))

    def log_prob(self, inputs, reduce=False, mask_index=None):
        if mask_index is not None:
            current_ordering = self.__mask_indicator
            self.reorder(mask_index=mask_index)
            results = self.density_estimator.log_prob(inputs=inputs, params_logits=self(inputs), reduce=reduce)
            self.reorder(current_ordering)
            return results

        results = 0.0
        for i in range(self.num_masks or 1):
            self.reorder()
            results += self.density_estimator.log_prob(inputs=inputs, params_logits=self(inputs), reduce=reduce)
        results /= self.num_masks or 1.0
        return results

    def sample(self, num_samples=1, mask_index=None):
        results = torch.rand(num_samples, self.in_features, device=self.ordering.device)
        is_training = self.training
        self.eval()
        if mask_index is not None:
            current_mask = self.__mask_indicator
            self.reorder(mask_index=mask_index)
        with torch.no_grad():
            for i in range(self.in_features):
                results = self.distributions(results).sample()
        if mask_index is not None:
            self.reorder(mask_index=current_mask)
        if is_training:
            self.train()
        return results

    def extra_repr(self):
        return "num_masks={}, seed={}".format(
            self.num_masks or "inf",
            self.seed,
        )



import torch
from pytorch_lightning.utilities.cli import LightningCLI
from mdade.training import MADETrainer
from mdade.data import DEDataModule
from mdade.__about__ import __version__


def main():
    print(f"mdade version {__version__} invoked")
    LightningCLI(MADETrainer, DEDataModule)


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages
from importlib.util import module_from_spec, spec_from_file_location
import typing as th
import re
import types
import os


# https://packaging.python.org/guides/single-sourcing-package-version/
# http://blog.ionelmc.ro/2014/05/25/python-packaging/
_PATH_ROOT = os.path.dirname(__file__)
_PATH_MADE_SRC = os.path.join(_PATH_ROOT, "mdade")
_PATH_REQUIREMENTS = os.path.join(_PATH_ROOT, "requirements")

# read description
with open(os.path.join(_PATH_ROOT, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# adapted from https://github.com/Lightning-AI/lightning/blob/master/setup.py
def _load_py_module(name: str, location: str) -> types.ModuleType:
    spec = spec_from_file_location(name, location)
    py = module_from_spec(spec)
    spec.loader.exec_module(py)
    return py


def _load_requirements(path_dir: str, file_name: str, comment_char: str = "#", unfreeze: bool = True) -> th.List[str]:
    """Load requirements from a file.
    >>> _load_requirements(_PATH_REQUIREMENTS)  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    ['numpy...', 'torch...', ...]
    """
    with open(os.path.join(path_dir, file_name)) as file:
        lines = [ln.strip() for ln in file.readlines()]
    reqs = []
    for ln in lines:
        # filer all comments
        comment = ""
        if comment_char in ln:
            comment = ln[ln.index(comment_char) :]
            ln = ln[: ln.index(comment_char)]
        req = ln.strip()
        # skip directly installed dependencies
        if not req or req.startswith("http") or "@http" in req:
            continue
        # remove version restrictions unless they are strict
        if unfreeze and "<" in req and "strict" not in comment:
            req = re.sub(r",? *<=? *[\d\.\*]+", "", req).strip()
        reqs.append(req)
    return reqs


_ABOUT_MODULE = _load_py_module(name="about", location=os.path.join(_PATH_MADE_SRC, "__about__.py"))

# https://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-extras
# Define package extras. These are only installed if you specify them.
# From remote, use like `pip install pytorch-lightning[dev, docs]`
# From local copy of repo, use like `pip install ".[dev, docs]"`
extras = {
    "tools": _load_requirements(path_dir=_PATH_REQUIREMENTS, file_name="tools.txt"),
}

setup(
    name="mdade",
    packages=find_packages(include=["*"]),
    version=_ABOUT_MODULE.__version__,
    description=_ABOUT_MODULE.__docs__,
    author=_ABOUT_MODULE.__author__,
    author_email=_ABOUT_MODULE.__author_email__,
    url=_ABOUT_MODULE.__homepage__,
    download_url="https://github.com/vahidzee/mdade",
    license=_ABOUT_MODULE.__license__,
    long_description=long_description,  # same as readme
    long_description_content_type="text/markdown",
    keywords=[
        "deep learning",
        "pytorch",
        "AI",
        "density estimation",
        "masked autoencoders",
        "pixelcnn",
        "autoregressive models",
    ],
    install_requires=_load_requirements(_PATH_REQUIREMENTS, "base.txt"),
    extras_require=extras,
    classifiers=[
        "Environment :: Console",
        "Natural Language :: English",
        "Development Status :: 3 - Alpha",
        # the project is intended for
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        # license
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        # Specify the Python versions you support here.
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    entry_points={
        "console_scripts": ["mdade=mdade.main:main"],
    },
)



import pytorch_lightning as pl
import typing as th
from torchvision.transforms import ToTensor
import torch
from torch.utils.data import Dataset, random_split, DataLoader
from .utils import initialize_transforms
from mdade.utils import get_value


class DEDataModule(pl.LightningDataModule):
    def __init__(
        self,
        # dataset
        dataset: th.Union[str, Dataset],
        dataset_args: th.Optional[dict] = None,
        train_dataset: th.Optional[th.Union[str, Dataset]] = None,
        train_dataset_args: th.Optional[dict] = None,
        val_size: th.Optional[th.Union[int, float]] = None,
        val_dataset: th.Optional[th.Union[str, Dataset]] = None,
        val_dataset_args: th.Optional[dict] = None,
        test_dataset: th.Optional[th.Union[str, Dataset]] = None,
        test_dataset_args: th.Optional[dict] = None,
        # transforms
        transforms: th.Optional[th.Union[dict, th.Any]] = None,
        train_transforms: th.Optional[th.Union[dict, th.Any]] = None,
        val_transforms: th.Optional[th.Union[dict, th.Any]] = None,
        test_transforms: th.Optional[th.Union[dict, th.Any]] = None,
        # batch_size
        batch_size: th.Optional[int] = 16,
        train_batch_size: th.Optional[int] = None,
        val_batch_size: th.Optional[int] = None,
        test_batch_size: th.Optional[int] = None,
        # shuffle
        train_shuffle: bool = True,
        val_shuffle: bool = False,
        test_shuffle: bool = False,
        # num_workers
        num_workers: int = 0,
        train_num_workers: th.Optional[int] = None,
        val_num_workers: th.Optional[int] = None,
        test_num_workers: th.Optional[int] = None,
        # seed
        seed: int = 0,
        # extra parameters (ignored)
        **kwargs,
    ):
        super().__init__()
        # transforms
        transforms = transforms if transforms is not None else ToTensor()
        self.transforms_train = initialize_transforms(
            train_transforms if train_transforms is not None else transforms
        )
        self.transforms_val = initialize_transforms(
            val_transforms if val_transforms is not None else transforms
        )
        self.transforms_test = initialize_transforms(
            test_transforms if test_transforms is not None else transforms
        )

        # seed
        self.seed = seed
        # normal classes

        # data
        self.train_data, self.val_data, self.test_data = None, None, None
        self.train_dataset = train_dataset if train_dataset is not None else dataset
        self.val_size = val_size
        assert (
            val_size is None
            or isinstance(val_size, int)
            or (isinstance(val_size, float) and val_size < 1)
        ), "invalid validation size is provided (either int, float (between zero and one) or None)"
        self.val_dataset = val_dataset if val_dataset is not None else dataset
        self.test_dataset = test_dataset if test_dataset is not None else dataset
        self.train_dataset_args = {
            **(dataset_args or dict()),
            **(train_dataset_args or dict()),
        }
        self.val_dataset_args = {
            **(dataset_args or dict()),
            **(val_dataset_args or dict()),
        }
        self.test_dataset_args = {
            **(dataset_args or dict()),
            **(test_dataset_args or dict()),
        }

        # batch_size
        self.train_batch_size = (
            train_batch_size if train_batch_size is not None else batch_size
        )
        self.val_batch_size = (
            val_batch_size if val_batch_size is not None else batch_size
        )
        self.test_batch_size = (
            test_batch_size if test_batch_size is not None else batch_size
        )
        assert (
            self.train_batch_size or self.val_batch_size or self.test_batch_size
        ), "at least one of batch_sizes should be a positive number"

        # shuffle
        self.train_shuffle, self.val_shuffle, self.test_shuffle = (
            train_shuffle,
            val_shuffle,
            test_shuffle,
        )

        # num_workers
        self.train_num_workers = (
            train_num_workers if train_num_workers is not None else num_workers
        )
        self.val_num_workers = (
            val_num_workers if val_num_workers is not None else num_workers
        )
        self.test_num_workers = (
            test_num_workers if test_num_workers is not None else num_workers
        )

    @staticmethod
    def get_dataset(dataset: th.Union[str, Dataset], **params):
        if isinstance(dataset, Dataset):
            return dataset
        else:
            return get_value(dataset, strict=True)(**params)

    def setup(self, stage: th.Optional[str] = None) -> None:
        if stage == "fit" and self.train_batch_size:
            dataset = self.get_dataset(
                self.train_dataset,
                train=True,
                transform=self.transforms_train,
                **self.train_dataset_args,
            )
            if not self.val_size:
                self.train_data = dataset
            if self.val_size and self.val_batch_size:
                train_len = (
                    len(dataset) - self.val_size
                    if isinstance(self.val_size, int)
                    else int(len(dataset) * (1 - self.val_size))
                )
                prng = torch.Generator()
                prng.manual_seed(self.seed)
                train_data, val_data = random_split(
                    dataset, [train_len, len(dataset) - train_len], generator=prng
                )
                self.train_data, self.val_data = train_data, val_data
            elif self.val_batch_size:
                self.val_data = self.get_dataset(
                    self.val_dataset,
                    transform=self.transforms_val,
                    **self.val_dataset_args,
                )

        elif stage == "test" and self.test_batch_size:
            self.test_data = (
                self.get_dataset(
                    self.test_dataset,
                    transform=self.transforms_test,
                    train=False,
                    **self.test_dataset_args,
                ),
            )

    def get_dataloader(
        self, name, batch_size=None, shuffle=None, num_workers=None, **params
    ):
        data = getattr(self, f"{name}_data")
        if data is None:
            return None
        batch_size = (
            batch_size
            if batch_size is not None
            else getattr(self, f"{name}_batch_size")
        )
        shuffle = shuffle if shuffle is not None else getattr(self, f"{name}_shuffle")
        num_workers = (
            num_workers
            if num_workers is not None
            else getattr(self, f"{name}_num_workers")
        )
        return DataLoader(
            dataset=data,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            **params,
        )

    def train_dataloader(self, **params):
        return self.get_dataloader("train", **params)

    def val_dataloader(self, **params):
        return self.get_dataloader("val", **params)

    def test_dataloader(self, **params):
        return self.get_dataloader("test", **params)



import time

_this_year = time.strftime("%Y")
__version__ = "0.0.4"
__author__ = "Vahid Zehtab"
__author_email__ = "vdzehtab@gmail.com"
__license__ = "MIT"
__copyright__ = f"Copyright (c) 2021-{_this_year}, {__author__}."
__homepage__ = "https://github.com/vahidzee/mdade"
# this has to be simple string, see: https://github.com/pypa/twine/issues/522
__docs__ = "'mdade' offers a lightweight PyTorch toolbox for Masked Autoregressive models"

__all__ = [
    "__author__",
    "__author_email__",
    "__copyright__",
    "__homepage__",
    "__license__",
    "__version__",
    "__docs__",
]



import torch
import typing as th
from .ordering_mixin import OrderedLayerMixin1D


class OrderedResidual1D(torch.nn.Module, OrderedLayerMixin1D):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        device: th.Optional[torch.device] = None,
        auto_connection: bool = True,
        scale=True,
    ) -> None:
        super().__init__()
        OrderedLayerMixin1D.__init__(
            self,
            in_features=in_features,
            out_features=out_features,
            device=device,
            mask_dtype=torch.float,
            auto_connection=auto_connection,
        )
        self.scale = scale

    def reorder(self, *args, **kwargs):
        OrderedLayerMixin1D.reorder(self, *args, **kwargs)
        if self.scale:
            mask = torch.div(self.mask, self.mask.sum(1, keepdim=True))
            self.mask.data.copy_(torch.where(mask != mask, torch.zeros_like(mask), mask))

    def forward(self, in_features, target_features) -> torch.Tensor:
        return target_features + in_features @ self.mask.T
